using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using System.IO;
using System.Diagnostics;
using nMail;

namespace csimap4
{
    public partial class FormMain : Form
    {
		public class TreeNodeData
		{
			public string Path;		// [{bNXtpX
			public char Separate;	// [{bNX؂蕶
			public int Flag;		// [{bNXtO
			public TreeNodeData(string path, char separate, int flag)
			{
				Path = path;
				Separate = separate;
				Flag = flag;
			}
		}

		// [{bNXc[̉摜ԍ
		enum IconImage {
			User,
			Inbox,
			Sent,
			Trash,
			Folder,
			NoSelect
		}
		// [T[o̐ݒ
		public class MailParameter
		{
			public string Smtp;
			public string Imap4;
			public string Address;
			public int SmtpPort;
			public int SmtpAuth;
			public int SmtpSsl;
			public int Imap4Port;
			public int Imap4Auth;
			public int Imap4Ssl;
		}
		private const int ContinueTimerCount = 60*10;
		private const string XmlFileName = "csimap4param.xml"; // [ݒۑt@C
		private string Id = "";				// IMAP4 ID
		private string Password = "";		// IMAP4 pX[h
		private string MailBox = "";		// ݑI𒆂̃[{bNX
		private string TrashMailBox = "";	// S~
		private string SentMailBox = "";	// M
		private bool FirstFlag = true;
		private bool DeleteFlag = false;	// [{bNXō폜H
		private MailParameter MailParam;	// [T[o̐ݒ
		private int TimerCount;
		private nMail.Imap4 imap;					// nMail.DLL IMAP4 NX

        public FormMain()
        {
            InitializeComponent();
        }

        private void FormMain_Load(object sender, EventArgs e)
        {
			MailParam = new MailParameter();
			MailParam.Imap4 = "";
			MailParam.Imap4Port = nMail.Imap4.StandardPortNo;
			MailParam.Imap4Ssl = 0;
			MailParam.Imap4Auth = nMail.Imap4.AuthLogin;
			MailParam.Smtp = "";
			MailParam.SmtpPort = nMail.Smtp.StandardPortNo;
			MailParam.SmtpSsl = 0;
			MailParam.SmtpAuth = 0;
			MailParam.Address = "";
			if(File.Exists(XmlFileName))
			{
				System.Xml.Serialization.XmlSerializer serializer = new System.Xml.Serialization.XmlSerializer(typeof(MailParameter));
				FileStream fs = new FileStream(XmlFileName, FileMode.Open);
				MailParam = (MailParameter)serializer.Deserialize(fs);
				fs.Close();
			} 
			// Winsock 
			Winsock.Initialize();
			Options.EnableConnectTimeout();
			// IMAP4
			imap = new nMail.Imap4();
			// UID gp
			imap.UseUid = true;
			// IDLE ݒ
			imap.IdleEvent += new System.EventHandler(ReceiveEvent);
			// [{bNX
			MailBox = "INBOX";
			TimerCount = 0;
        }

		private delegate void ReceiveEventDelegate(object sender, System.EventArgs e);

		// [MCxg
		private void ReceiveEvent(object sender, System.EventArgs e)
		{
			if(InvokeRequired)
			{
				BeginInvoke(new ReceiveEventDelegate(ReceiveEvent), new object[] { sender, e });
			}
			else
			{
				GetMailList();
				imap.StartIdle();
			}
		}

		// IMAP4 F
		// łɐڑς݂Ȃ炻̂܂ܖ߂
		private bool Authenticate()
		{
			TimerCount = ContinueTimerCount;
			// ڑH
			if(!imap.IsConnected()) {
				try
				{
					imap.SSL = MailParam.Imap4Ssl;
					// ڑ
					imap.Connect(MailParam.Imap4, MailParam.Imap4Port);
					// F
					imap.Authenticate(Id, Password, MailParam.Imap4Auth);
					return true;
				}
				catch(nMail.nMailException nex)
				{
					if(nex.ErrorCode == nMail.Imap4.ErrorAuthenticate) 
					{
						MessageBox.Show(string.Format("F؃G[łB\r\n{0}", nex.Message));
						Id = "";
						Password = "";
					}
					else
					{
						MessageBox.Show(string.Format("T[o[ڑɃG[܂B\r\n{0}", nex.Message));
					}
				}
				catch(Exception ex)
				{
					Debug.WriteLine(ex.Message);
				}
				// G[Ȃ
				imap.Close();
				return false;
			}
			return true;
		}

		//	̂̃[{bNXo^ς݂H
		private TreeNode SearchNode(TreeNode node, string path)
		{
			foreach(TreeNode tn in node.Nodes)
			{
				TreeNode ctn;
				if((ctn = SearchNode(tn, path)) != null)
				{
					return ctn;
				}
				if(tn.Tag != null) 
				{
					TreeNodeData data = (TreeNodeData)tn.Tag;
					if(string.Compare(data.Path, path, true) == 0)
					{
						return tn;
					}
				}
			}
			return null;
		}

		// [{bNXc[ɒǉ
		private void AddTree(string name, char separate, int flag)
		{
			int no;
			TreeNode node, parent_node;
			string path;
			string [] folder = name.Split(separate);

			parent_node = treeMailBox.TopNode;
			treeMailBox.PathSeparator = separate.ToString();
			for(no = 0 ; no < folder.Length ; no++)
			{
				path = string.Join(separate.ToString(), folder, 0, no + 1);
				node = SearchNode(treeMailBox.TopNode, path);
				if(node == null)
				{
					int index;
					string text = folder[no];
					index = (int)IconImage.Folder;
					// IłȂ[{bNX
					if((flag & nMail.Imap4.MailBoxNoSelect) != 0)
					{
						index = (int)IconImage.NoSelect;
					}
					if(no == 0)
					{
						if(string.Compare(text, "inbox", true) == 0)
						{
							text = "M";
							index = (int)IconImage.Inbox;
						}
						else if(string.Compare(text, "sent", true) == 0)
						{
							SentMailBox = text;
							text = "M";
							index = (int)IconImage.Sent;
						}
						else if(string.Compare(text, "trash", true) == 0)
						{
							TrashMailBox = text;
							text = "S~";
							index = (int)IconImage.Trash;
						}
					}
					int insert_no = 0;
					foreach(TreeNode tn in parent_node.Nodes)
					{
						if(index < tn.ImageIndex)
						{
							break;
						}
						insert_no++;
					}
					node = new TreeNode(text, index, index);
					parent_node.Nodes.Insert(insert_no, node);
					node.Tag = new TreeNodeData(path, separate, flag);
					parent_node.Expand();
				}
				parent_node = node;
			}
		}

		// [{bNXc[r[ɕ\
		private bool GetMailBoxTree()
		{
			treeMailBox.Nodes.Clear();
			if(Authenticate())
			{
				try 
				{
					TreeNode add_node = treeMailBox.Nodes.Add(MailParam.Address);
					add_node.ImageIndex = (int)IconImage.User;
					// [{bNXꗗ
					imap.GetMailBoxList("", "*");
					nMail.Imap4.MailBoxStatus [] mailbox_list = imap.GetMailBoxStatusList();
					foreach(nMail.Imap4.MailBoxStatus mailbox in mailbox_list)
					{
						if(treeMailBox.TopNode.Tag == null) {
							treeMailBox.TopNode.Tag = new TreeNodeData("", mailbox.Separate, 0);
						}
						// c[r[ɍڒǉ
						AddTree(mailbox.Name, mailbox.Separate, mailbox.Flag);
					}
					// INBOX I
					treeMailBox.SelectedNode = SearchNode(treeMailBox.TopNode, MailBox);
					return true;
				}
				catch(nMail.nMailException nex)
				{
					MessageBox.Show(string.Format("G[܂B\r\n{0}", nex.Message));
				}
				catch(Exception ex)
				{
					Debug.WriteLine(ex.Message);
				}
			}
			return false;
		}

		// ݑI𒆂̃[{bNX̃[ꗗ\
		private void GetMailList()
		{
			// ID ͂Ȃ̓_CAO\
			if(Id == "") 
			{
				if(!InputId()) 
				{
					return;
				}
			}
			// [{bNXꗗ܂\ĂȂ
			// 擾ĕ\
			if(treeMailBox.TopNode == null) 
			{
				if(!GetMailBoxTree())
				{
					return;
				}
			}
			textBody.Text = "";
			listMail.Items.Clear();
			listAttach.Visible = false;
			if(MailBox != "")
			{
				if(Authenticate())
				{
					try 
					{
						if(DeleteFlag)
						{
							// 폜tÔ[폜
							imap.Expunge();
							DeleteFlag = false;
						}
						// [{bNXI
						imap.SelectMailBox(MailBox);
						// [{bNX̍폜}[NȂ[ UID 擾
						imap.Search(nMail.Imap4.SearchUnDeleted);
						uint [] result_list = imap.GetSearchMailResultList();
						foreach(uint uid in result_list)
						{
							// [̃Xe[^X擾
							imap.GetStatus(uid);
							// .NET Framework 2.0 ȍ~łȂƃI[i[h[ʓ|Ȃ悤Ȃ̂ŁA
							// Ƃ肠ǂ""ƕ\邱Ƃ
							string seen = "";
							if((imap.MessageFlag & nMail.Imap4.MessageSeen) == 0)
							{
								seen = "";
							}
							string attach = "";
							if(imap.GetAttachmentFileStatus(uid)) 
							{
								attach = "Y";
							}
							string[] item = { seen, attach, imap.Subject, imap.From, imap.DateString };
							ListViewItem add_item = listMail.Items.Add(new ListViewItem(item));
							// UID ACeɕۑ
							add_item.Tag = uid;
						}
					}
					catch(nMail.nMailException nex)
					{
						MessageBox.Show(string.Format("G[܂B\r\n{0}", nex.Message));
					}
					catch(Exception ex)
					{
						Debug.WriteLine(ex.Message);
					}
				}
			}
		}

		// ID ƃpX[h
		private bool InputId()
		{
			FormId form = new FormId();

			form.Id = Id;
			form.Password = Password;
			if(form.ShowDialog() == DialogResult.OK) 
			{
				Id = form.Id;
				Password = form.Password;
				if(Id != "" && Password != "") 
				{
					return true;
				}
			}
			return false;
		}

        private void FormMain_Activated(object sender, EventArgs e)
        {
			if(FirstFlag) 
			{
				FirstFlag = false;
				if(MailParam.Imap4 == "" || MailParam.Imap4Port == 0) 
				{
					menuOption_Click(sender, e);
				}
				if(MailParam.Imap4 != "") 
				{
					GetMailList();
					imap.StartIdle();
				}
			}
        }

		// gqɑΉACR擾p API ֘A̐錾
		[DllImport("shell32.dll")]
		private static extern IntPtr SHGetFileInfo(string pszPath, uint dwFileAttributes, ref SHFILEINFO psfi, uint cbSizeFileInfo, uint uFlags);
		private const uint SHGFI_ICON = 0x0100;
		private const uint SHGFI_SMALLICON = 0x0001;
		private const uint SHGFI_USEFILEATTRIBUTES = 0x0010;
		private struct SHFILEINFO
		{
			public IntPtr hIcon;
			public IntPtr iIcon;
			public uint dwAttributes;
			[MarshalAs(UnmanagedType.ByValTStr, SizeConst = 260)]
			public string szDisplayName;
			[MarshalAs(UnmanagedType.ByValTStr, SizeConst = 80)]
			public string szTypeName;
		}; 

        private void listMail_SelectedIndexChanged(object sender, EventArgs e)
        {
			if(listMail.SelectedItems.Count > 0)
			{
				if(Authenticate()) 
				{
					try 
					{
						textBody.Text = "";
						// UID 擾
						uint uid = (uint)listMail.SelectedItems[0].Tag;
						int count = 0;
						// Ytt@C\Xgr[NA
						listAttach.Items.Clear();
						// Ytt@C̃ACRpC[WXgNA
						imageAttach.Images.Clear();
						// MIME \擾
						imap.GetMimeStructure(uid);
						nMail.Imap4.MimeStructureStatus [] mime_list = imap.GetMimeStructureStatusList();
						foreach(nMail.Imap4.MimeStructureStatus mime in mime_list)
						{
							// eLXgłΖ{ɕ\
							if(string.Compare(mime.Type, "text", true) == 0 && string.Compare(mime.SubType, "plain", true) == 0)
							{
								imap.GetMimePart(uid, mime.PartNo);
								textBody.Text = imap.Body;
							}
							else if(string.Compare(mime.Type, "multipart", true) != 0)
							{
								// Ytt@Cł΁At@CƃACR
								// Ytt@C\Xgr[ɕ\
								string name = mime.FileName;
								// HTML [̖{̂ noname.html ɂȂ̂ŁA
								// Ƃ肠킩₷ȖOɕύX
								if(name == "noname.html") 
								{
									name = "{.html";
								}
								// gqɑΉACR̎擾
								SHFILEINFO sf = new SHFILEINFO();
								SHGetFileInfo(name, 0, ref sf, (uint)Marshal.SizeOf(sf), SHGFI_ICON | SHGFI_SMALLICON | SHGFI_USEFILEATTRIBUTES);
								Icon icon = Icon.FromHandle(sf.hIcon);
								// Ytt@C\Xgr[ɒǉ
								imageAttach.Images.Add(icon);
								ListViewItem add_item = listAttach.Items.Add(name, count++);
								add_item.Tag = mime.PartNo;
							}
						}
						// Ytt@C΃Xgr[\
						listAttach.Visible = false;
						if(listAttach.Items.Count > 0)
						{
							listAttach.Visible = true;
						}
						listMail.SelectedItems[0].SubItems[0].Text = "";

						imap.StartIdle();
					}
					catch(nMail.nMailException nex)
					{
						MessageBox.Show(string.Format("G[܂B\r\n{0}", nex.Message));
					}
					catch(Exception ex)
					{
						Debug.WriteLine(ex.Message);
					}
				}
			}
        }

        private void listAttach_DoubleClick(object sender, EventArgs e)
        {
			menuOpenAttach_Click(sender, e);
        }

        private void treeMailBox_MouseDown(object sender, MouseEventArgs e)
        {
			// ENbN̍ۂɃ[{bNX̑IύX
			if(e.Button == MouseButtons.Right)
			{
				Point point = treeMailBox.PointToClient(Cursor.Position);
				treeMailBox.SelectedNode = treeMailBox.GetNodeAt(point);
			}
        }

		// q[{bNX̖OύX
		private void ChangeChildString(TreeNode node, string old_mailbox, string new_mailbox)
		{
			foreach(TreeNode tn in node.Nodes)
			{
				if(tn.Tag != null) 
				{
					TreeNodeData data = (TreeNodeData)tn.Tag;
					if(data.Path.StartsWith(old_mailbox))
					{
						string path = new_mailbox + data.Path.Substring(old_mailbox.Length);
						tn.Tag = new TreeNodeData(path, data.Separate, data.Flag);
					}
				}
				ChangeChildString(tn, old_mailbox, new_mailbox);
			}
		}

        private void treeMailBox_AfterLabelEdit(object sender, NodeLabelEditEventArgs e)
        {
			if(e.Label != null && e.Label != e.Node.Text)
			{
				string new_mailbox, old_mailbox;
				TreeNodeData data = (TreeNodeData)e.Node.Tag;
				// [{bNX擾
				old_mailbox = data.Path;
				// V[{bNX쐬
				string [] folder = data.Path.Split(data.Separate);
				new_mailbox = "";
				if(folder.Length > 1)
				{
					new_mailbox = string.Join(data.Separate.ToString(), folder, 0, folder.Length - 1);
					new_mailbox += data.Separate.ToString();
				}
				new_mailbox += e.Label;
				if(Authenticate()) 
				{
					try 
					{
						// [{bNXύX
						imap.RenameMailBox(old_mailbox, new_mailbox);
						// [{bNXXV
						e.Node.Tag = new TreeNodeData(new_mailbox, data.Separate, data.Flag);
						// q[{bNX̍XV
						ChangeChildString(e.Node, old_mailbox, new_mailbox);
						imap.StartIdle();
						return;
					}
					catch(nMail.nMailException nex)
					{
						MessageBox.Show(string.Format("G[܂B\r\n{0}", nex.Message));
					}
					catch(Exception ex)
					{
						Debug.WriteLine(ex.Message);
					}
					e.CancelEdit = true;
				}
			}
			else
			{
				e.CancelEdit = true;
			}
        }

		// [{bNX̑IύX
        private void treeMailBox_AfterSelect(object sender, TreeViewEventArgs e)
        {
			if(treeMailBox.SelectedNode.Tag != null)
			{
				TreeNodeData data = (TreeNodeData)treeMailBox.SelectedNode.Tag;
				if((data.Flag & nMail.Imap4.MailBoxNoSelect) == 0)
				{
					MailBox = data.Path;
					GetMailList();
					imap.StartIdle();
				}
			}
        }

		// [{bNXOύXJnȌ
        private void treeMailBox_BeforeLabelEdit(object sender, NodeLabelEditEventArgs e)
        {
			// [AhXƎMAMρAS~͕ύXs
			if(e.Node.ImageIndex != (int)IconImage.User && e.Node.ImageIndex < (int)IconImage.Folder) 
			{
				e.CancelEdit = true;
			}
        }

        private void treeMailBox_BeforeSelect(object sender, TreeViewCancelEventArgs e)
        {
			if(e.Node != null)
			{
				TreeNodeData data = (TreeNodeData)e.Node.Tag;
				if(data != null && (data.Flag & nMail.Imap4.MailBoxNoSelect) != 0)
				{
					e.Cancel = true;
				}
			}
        }

        private void treeMailBox_KeyUp(object sender, KeyEventArgs e)
        {
			// F2 L[Ń[{bNXOύXJn
			if(e.KeyCode == Keys.F2 && treeMailBox.SelectedNode != null) 
			{
				treeMailBox.SelectedNode.BeginEdit();
			}
        }

		// [{bNXc[r[Ńhbvꂽ
        private void treeMailBox_DragDrop(object sender, DragEventArgs e)
        {
			if(e.Data.GetDataPresent(typeof(ListViewItem)))
			{
				e.Effect = DragDropEffects.None;
				Point point = treeMailBox.PointToClient(new Point(e.X, e.Y));
				TreeNode node = treeMailBox.GetNodeAt(point);
				if(node != null)
				{
					TreeNodeData data = (TreeNodeData)node.Tag;
					if(node.ImageIndex != (int)IconImage.User && (data.Flag & nMail.Imap4.MailBoxNoSelect) == 0)
					{
						ListViewItem item = (ListViewItem)e.Data.GetData(typeof(ListViewItem));
						if(Authenticate())
						{
							try 
							{
								// hbvꂽ[hbv[{bNXɃRs[
								uint uid = (uint)item.Tag;
								imap.Copy(uid, data.Path);
								e.Effect = DragDropEffects.Move;

								imap.StartIdle();
							}
							catch(nMail.nMailException nex)
							{
								MessageBox.Show(string.Format("G[܂B\r\n{0}", nex.Message));
							}
							catch(Exception ex)
							{
								Debug.WriteLine(ex.Message);
							}
						}
					}
				}
			}
        }

		// [{bNXc[r[փhbOڂ
        private void treeMailBox_DragEnter(object sender, DragEventArgs e)
        {
			e.Effect = DragDropEffects.None;	
			if(e.Data.GetDataPresent(typeof(ListViewItem)))
			{
				Point point = treeMailBox.PointToClient(new Point(e.X, e.Y));
				TreeNode node = treeMailBox.GetNodeAt(point);
				if(node != null)
				{
					TreeNodeData data = (TreeNodeData)node.Tag;
					if(node.ImageIndex != (int)IconImage.User && (data.Flag & nMail.Imap4.MailBoxNoSelect) == 0)
					{
						e.Effect = DragDropEffects.Move;
					}
				}
			}
        }

		// [{bNXc[r[ŃhbOڂړ
        private void treeMailBox_DragOver(object sender, DragEventArgs e)
        {
			e.Effect = DragDropEffects.None;	
			if(e.Data.GetDataPresent(typeof(ListViewItem)))
			{
				Point point = treeMailBox.PointToClient(new Point(e.X, e.Y));
				TreeNode node = treeMailBox.GetNodeAt(point);
				if(node != null)
				{
					TreeNodeData data = (TreeNodeData)node.Tag;
					if(node.ImageIndex != (int)IconImage.User && (data.Flag & nMail.Imap4.MailBoxNoSelect) == 0)
					{
						e.Effect = DragDropEffects.Move;
					}
				}
			}
        }

        private void menuCreate_Click(object sender, EventArgs e)
        {
			// [̍쐬
			FormWrite form = new FormWrite();

			form.HostName = MailParam.Smtp;
			form.From = MailParam.Address;
			form.SSL = MailParam.SmtpSsl;
			form.PortNo = MailParam.SmtpPort;
			form.Auth = MailParam.SmtpAuth;
			form.Id = Id;
			form.Password = Password;
			if(form.ShowDialog() == DialogResult.OK)
			{
				// Mσ[{bNXΑMς݂̃[Abv[h
				AppendSent(form.Subject, MailParam.Address, form.To, form.Body, form.FileName);
			}
        }

		// [ꗗXV
        private void menuRedisp_Click(object sender, EventArgs e)
        {
			GetMailList();
			imap.StartIdle();
        }

		// I
        private void menuExit_Click(object sender, EventArgs e)
        {
			Close();
        }

		// [T[o̐ݒ
        private void menuOption_Click(object sender, EventArgs e)
        {
			FormOption form = new FormOption();

			form.Imap4 = MailParam.Imap4;
			form.Imap4Port = MailParam.Imap4Port;
			form.Imap4Ssl = MailParam.Imap4Ssl;
			form.Imap4Auth = MailParam.Imap4Auth;
			form.Smtp = MailParam.Smtp;
			form.SmtpPort = MailParam.SmtpPort;
			form.SmtpSsl = MailParam.SmtpSsl;
			form.SmtpAuth = MailParam.SmtpAuth;
			form.Address = MailParam.Address;
			if(form.ShowDialog() == DialogResult.OK) 
			{
				MailParam.Imap4 = form.Imap4;
				MailParam.Imap4Port = form.Imap4Port;
				MailParam.Imap4Ssl = form.Imap4Ssl;
				MailParam.Imap4Auth = form.Imap4Auth;
				MailParam.Smtp = form.Smtp;
				MailParam.SmtpPort = form.SmtpPort;
				MailParam.SmtpSsl = form.SmtpSsl;
				MailParam.SmtpAuth = form.SmtpAuth;
				MailParam.Address = form.Address;

				System.Xml.Serialization.XmlSerializer serializer = new System.Xml.Serialization.XmlSerializer(typeof(MailParameter));
				FileStream fs = new FileStream(XmlFileName, FileMode.Create);
				serializer.Serialize(fs, MailParam);
				fs.Close();
			}
        }

		// [{bNX쐬
        private void menuCreateMailBox_Click(object sender, EventArgs e)
        {
			if(treeMailBox.SelectedNode != null)
			{
				TreeNodeData data = (TreeNodeData)treeMailBox.SelectedNode.Tag;
				// q[{bNX쐬ł邩H
				if((data.Flag & nMail.Imap4.MailBoxNoInferious) == 0)
				{
					FormName form = new FormName();

					// [{bNX̓_CAO\
					if(form.ShowDialog() == DialogResult.OK) 
					{
						if(Authenticate())
						{
							try 
							{
								string path = "";
								if(data.Path != "") {
								 	path = data.Path + data.Separate;
								}
								path += form.MailBox;
								// [{bNX쐬
								imap.CreateMailBox(path);
								// w
								imap.SubscribeMailBox(path);
								int index;
								string text;

								index = (int)IconImage.Folder;
								text = form.MailBox;
								// sent ̏ꍇMσ[{bNXƂĈ
								if(string.Compare(path, "sent", true) == 0)
								{
									SentMailBox = path;
									text = "M";
									index = (int)IconImage.Sent;
								}
								// trash ̏ꍇS~ƂĈ
								else if(string.Compare(text, "trash", true) == 0)
								{
									TrashMailBox = path;
									text = "S~";
									index = (int)IconImage.Trash;
								}
								TreeNode add_node = treeMailBox.SelectedNode.Nodes.Add(text);
								add_node.ImageIndex = index;
								add_node.SelectedImageIndex = index;
								add_node.Tag = new TreeNodeData(path, data.Separate, 0);
								treeMailBox.SelectedNode.Expand();
								imap.StartIdle();
							}
							catch(nMail.nMailException nex)
							{
								MessageBox.Show(string.Format("G[܂B\r\n{0}", nex.Message));
							}
							catch(Exception ex)
							{
								Debug.WriteLine(ex.Message);
							}
						}
					}
				}
			}
        }

		// [{bNX̖OύX
        private void menuRenameMailBox_Click(object sender, EventArgs e)
        {
			if(treeMailBox.SelectedNode != null) 
			{
				treeMailBox.SelectedNode.BeginEdit();
			}
        }

		// q[{bNX̍폜
		private void DeleteChildMailBox(TreeNode node)
		{
			foreach(TreeNode tn in node.Nodes)
			{
				DeleteChildMailBox(tn);
				if(tn.Tag != null) 
				{
					TreeNodeData data = (TreeNodeData)tn.Tag;
					// [{bNX폜
					imap.DeleteMailBox(data.Path);
					// wǉ
					imap.UnsubscribeMailBox(data.Path);
					imap.StartIdle();
				}
			}
		}

		// [{bNX̍폜
        private void menuDeleteMailBox_Click(object sender, EventArgs e)
        {
			if(Authenticate())
			{
				try 
				{
					// MAMρAS~͍폜łȂ
					if(treeMailBox.SelectedNode.ImageIndex >= (int)IconImage.Folder)
					{
						// q[{bNXSč폜
						DeleteChildMailBox(treeMailBox.SelectedNode);
						if(treeMailBox.SelectedNode.Tag != null)
						{
							// [{bNX폜
							TreeNodeData data = (TreeNodeData)treeMailBox.SelectedNode.Tag;
							imap.DeleteMailBox(data.Path);
							// wǉ
							imap.UnsubscribeMailBox(data.Path);
							treeMailBox.SelectedNode.Remove();
							imap.StartIdle();
						}
					}
				}
				catch(nMail.nMailException nex)
				{
					MessageBox.Show(string.Format("G[܂B\r\n{0}", nex.Message));
				}
				catch(Exception ex)
				{
					Debug.WriteLine(ex.Message);
				}
			}
        }

		// Mσ[{bNXΑMς݂̃[Abv[h
		private void AppendSent(string subject, string from, string to, string body, string filename)
		{
			// Mσ[{bNX邩H
			if(SentMailBox != "")
			{
				if(Authenticate())
				{
					try 
					{
						string header;
						header = "Subject: " + subject + "\r\n";
						header += "From: " + from + "\r\n";
						header += "To: " + to + "\r\n";
						imap.AppendMail(SentMailBox, header, body, nMail.Imap4.MessageSeen, filename);
						imap.StartIdle();
					}
					catch(nMail.nMailException nex)
					{
						MessageBox.Show(string.Format("G[܂B\r\n{0}", nex.Message));
					}
					catch(Exception ex)
					{
						Debug.WriteLine(ex.Message);
					}
				}
			}
		}

		// ԐM
        private void menuReply_Click(object sender, EventArgs e)
        {
			if(listMail.SelectedItems.Count > 0) 
			{
				FormWrite form = new FormWrite();

				form.HostName = MailParam.Smtp;
				form.From = MailParam.Address;
				form.Subject = "Re:" + listMail.SelectedItems[0].SubItems[1].Text;
				form.To = listMail.SelectedItems[0].SubItems[2].Text;
				form.SSL = MailParam.SmtpSsl;
				form.PortNo = MailParam.SmtpPort;
				form.Auth = MailParam.SmtpAuth;
				form.Id = Id;
				form.Password = Password;
				if(form.ShowDialog() == DialogResult.OK)
				{
					// Mσ[{bNXΑMς݂̃[Abv[h
					AppendSent(form.Subject, MailParam.Address, form.To, form.Body, form.FileName);
				}
			}
        }

		// [̍폜
        private void menuDeleteMail_Click(object sender, EventArgs e)
        {
			if(Authenticate())
			{
				try 
				{
					ListViewItem item = listMail.SelectedItems[0];
					uint uid = (uint)item.Tag;
					if(TrashMailBox != "" && MailBox != TrashMailBox)
					{
						imap.Copy(uid, TrashMailBox);
					}
					imap.Delete(uid);
					listMail.Items.Remove(listMail.SelectedItems[0]);
					DeleteFlag = true;
					imap.StartIdle();
				}
				catch(nMail.nMailException nex)
				{
					MessageBox.Show(string.Format("G[܂B\r\n{0}", nex.Message));
				}
				catch(Exception ex)
				{
					Debug.WriteLine(ex.Message);
				}
			}
        }

		// [vpeB\
        private void menuProperty_Click(object sender, EventArgs e)
        {
			if(Authenticate()) 
			{
				try 
				{
					// UID 擾
					uint uid = (uint)listMail.SelectedItems[0].Tag;
					imap.GetStatus(uid);
					// MIME \擾
					imap.GetMimeStructure(uid);
					FormProperty form = new FormProperty();
					form.Header = imap.Header;
					nMail.Imap4.MimeStructureStatus [] mime_list = imap.GetMimeStructureStatusList();
					// MIME \̒ǉ
					foreach(nMail.Imap4.MimeStructureStatus mime in mime_list)
					{
						string text;
						text = mime.Type + "/" + mime.SubType + "; ";
						for(int no = 0 ; no < mime.Parameter.Length ; no++)
						{
							text += mime.Parameter[no].Name + "=" + mime.Parameter[no].Value;
							if(no < mime.Parameter.Length - 1)
							{
								text += "; ";
							}
						}
						form.AddTreeItem(mime.Part, text);
					}
					imap.StartIdle();
					form.ShowDialog();
				}
				catch(nMail.nMailException nex)
				{
					MessageBox.Show(string.Format("G[܂B\r\n{0}", nex.Message));
				}
				catch(Exception ex)
				{
					Debug.WriteLine(ex.Message);
				}
			}
        }

		// Ytt@CJ
        private void menuOpenAttach_Click(object sender, EventArgs e)
        {
			if(Authenticate()) 
			{
				try 
				{
					uint uid = (uint)listMail.SelectedItems[0].Tag;
					int part_no = (int)listAttach.SelectedItems[0].Tag;
					string path = Path.GetTempPath() + listAttach.SelectedItems[0].Text;
					// Ytt@Ce|tH_ɕۑ
					imap.SaveMimePart(uid, part_no, path);
					// ۑt@CJ
					Process.Start(path);
					imap.StartIdle();
				}
				catch(nMail.nMailException nex)
				{
					MessageBox.Show(string.Format("G[܂B\r\n{0}", nex.Message));
				}
				catch(Exception ex)
				{
					Debug.WriteLine(ex.Message);
				}
			}
        }

		// Ytt@Cۑ
        private void menuSaveAttach_Click(object sender, EventArgs e)
        {
			if(Authenticate()) 
			{
				try 
				{
					uint uid = (uint)listMail.SelectedItems[0].Tag;
					int part_no = (int)listAttach.SelectedItems[0].Tag;
					saveFileDialog.FileName = listAttach.SelectedItems[0].Text;
					// t@Cۑ_CAO\
					if(saveFileDialog.ShowDialog() == DialogResult.OK)
					{
						// Ytt@Cۑ
						imap.SaveMimePart(uid, part_no, saveFileDialog.FileName);
						imap.StartIdle();
					}
				}
				catch(nMail.nMailException nex)
				{
					MessageBox.Show(string.Format("G[܂B\r\n{0}", nex.Message));
				}
				catch(Exception ex)
				{
					Debug.WriteLine(ex.Message);
				}
			}
        }

		// [{bNXc[r[̃ReLXgj[\Ȍ
        private void contextTree_Popup(object sender, EventArgs e)
        {
			bool flag = false;
			TreeNode node = treeMailBox.SelectedNode;
			if(node != null)
			{
				if(node.ImageIndex >= (int)IconImage.Folder) 
				{
					flag = true;
				}
				TreeNodeData data = (TreeNodeData)treeMailBox.SelectedNode.Tag;
				// q[{bNX쐬ł邩H
				if((data.Flag & nMail.Imap4.MailBoxNoInferious) == 0)
				{
					menuCreateMailBox.Enabled = true;
				}
			} 
			else
			{
				menuCreateMailBox.Enabled = false;
			}
			menuDeleteMailBox.Enabled = flag;
			menuRenameMailBox.Enabled = flag;
        }

		// [ꗗł̃ReLXgj[\Ȍ
        private void contextMail_Popup(object sender, EventArgs e)
        {
			// Iꂽ[Ȃ΍폜͕s
			if(listMail.SelectedItems.Count == 0)
			{
				menuDeleteMail.Enabled = false;
				menuReply.Enabled = false;
				menuProperty.Enabled = false;
			}
			else
			{
				menuDeleteMail.Enabled = true;
				menuReply.Enabled = true;
				menuProperty.Enabled = true;
			}
        }

        private void timer_Tick(object sender, EventArgs e)
        {
			if(TimerCount > 0) 
			{
				TimerCount--;
				if(TimerCount == 0 && imap.IsConnected()) 
				{
					// 莞ԒʐM̏ꍇ No Operation R}h𔭍s
					try
					{
						imap.NoOperation();
						TimerCount = ContinueTimerCount;
						imap.StartIdle();
					}
					catch(nMail.nMailException nex)
					{
						Debug.WriteLine(nex.Message);
						imap.Close();
					}
					catch(Exception ex)
					{
						Debug.WriteLine(ex.Message);
					}
				}
			}
        }

        private void FormMain_FormClosed(object sender, FormClosedEventArgs e)
        {
			// 폜tÔ[폜
			if(DeleteFlag)
			{
				imap.Expunge();
			}
			// ڑI
			imap.Close();
			// Winsock I
			Winsock.Done();
        }

		// [ꗗ̃hbOhbvJn
        private void listMail_ItemDrag(object sender, ItemDragEventArgs e)
        {
			if(e.Item != null)
			{
				if(listMail.DoDragDrop((ListViewItem)e.Item, DragDropEffects.Move) == DragDropEffects.Move)
				{
					if(Authenticate())
					{
						try 
						{
							// hbO[폜
							ListViewItem item = (ListViewItem)e.Item;
							uint uid = (uint)item.Tag;
							imap.Delete(uid);
							listMail.Items.Remove(item);
							DeleteFlag = true;
						}
						catch(nMail.nMailException nex)
						{
							MessageBox.Show(string.Format("G[܂B\r\n{0}", nex.Message));
						}
						catch(Exception ex)
						{
							Debug.WriteLine(ex.Message);
						}
					}
				}
			}
        }
    }
}